@@ -1,6 +1,6 @@
 #!/usr/bin/python
 # -*- coding: utf-8 -*-
-# Copyright (C) 2012-17 Dr. Ralf Schlatterbeck Open Source Consulting.
+# Copyright (C) 2012-21 Dr. Ralf Schlatterbeck Open Source Consulting.
 # Reichergasse 131, A-3411 Weidling.
 # Web: http://www.runtux.com Email: office@runtux.com
 # All rights reserved
@@ -30,7 +30,7 @@ import csv
 from   datetime           import datetime, time, tzinfo, timedelta
 from   rsclib.stateparser import Parser, Parse_Error
 from   rsclib.autosuper   import autosuper
-from   rsclib.pycompat    import ustr
+from   rsclib.pycompat    import ustr, string_types
 
 class TZ (tzinfo) :
     def __init__ (self, offset = 0) :
@@ -95,15 +95,16 @@ class SQL_Type (autosuper) :
         # decimal dot if the microsecond part was all zero to make it
         # roundtrip
         if getattr (self, 'dtfmt', None) :
-            return value.strftime (self.dtfmt)
+            return value.strftime (self.dtfmt).encode ('ascii')
         if getattr (self, 'timefmt', None) :
-            return value.strftime (self.timefmt).rstrip (b'0').rstrip (b'.')
-        return repr (value)
+            v = value.strftime (self.timefmt).encode ('ascii')
+            return v.rstrip (b'0').rstrip (b'.')
+        return repr (value).encode ('ascii')
     # end def format
 
     def typ (self, tn) :
         if self.parameters :
-            return "%s(%s)" % (tn, ','.join (self.parameters))
+            return b"%s(%s)" % (tn, b','.join (self.parameters))
         return tn
     # end def typ
 
@@ -139,17 +140,27 @@ class SQL_boolean (SQL_Type) :
 class SQL_double (SQL_Type) :
 
     def __call__ (self, f) :
-        if f == '\\N' or f == 'NULL' :
+        if f == b'\\N' or f == b'NULL' :
             return None
         return float (f)
     # end def __call__
 
 # end class SQL_double
 
+class SQL_real (SQL_Type) :
+
+    def __call__ (self, f) :
+        if f == b'\\N' or f == b'NULL' :
+            return None
+        return float (f)
+    # end def __call__
+
+# end class SQL_real
+
 class SQL_integer (SQL_Type) :
 
     def __call__ (self, i) :
-        if i == '\\N' or i == 'NULL' :
+        if i == b'\\N' or i == b'NULL' :
             return None
         return int (i)
     # end def __call__
@@ -165,9 +176,9 @@ class SQL_numeric (SQL_Type) :
     # end def __init__
 
     def __call__ (self, n) :
-        if n == '\\N' or n == 'NULL' :
+        if n == b'\\N' or n == b'NULL' :
             return None
-        a, b = n.split ('.')
+        a, b = n.split (b'.')
         return (int (a), int (b))
     # end def __call__
 
@@ -328,7 +339,7 @@ class SQL_character (SQL_Type) :
     def format (self, dialect, typ, value) :
         if value is None :
             return dialect.sql_null
-        return value.encode ('utf-8')
+        return value
     # end def format
 
 # end class SQL_character
@@ -374,6 +385,8 @@ class SQL_Timestamp_Without_Zone (SQL_Type) :
     timefmt = '%Y-%m-%d %H:%M:%S.%f'
 
     def __call__ (self, ts) :
+        if isinstance (ts, type (b'')) :
+            ts = ts.decode ('ascii')
         if ts == '\\N' or ts == 'NULL' or ts == '0000-00-00 00:00:00' :
             return None
         fmt = self.timefmt
@@ -438,6 +451,9 @@ def make_naive (dt) :
 # end make_naive
 
 class adict (dict) :
+    """ A dictionary that is a little more tolerant *and* is able to
+        access elements with an attribute access.
+    """
 
     def __init__ (self, *args, **kw) :
         self.done = False
@@ -445,11 +461,24 @@ class adict (dict) :
     # end def __init__
 
     def __getattr__ (self, key) :
-        if key in self :
-            return self [key]
+        if key.startswith ('public') :
+            key = key.split ('.', 1) [-1]
+        for k in key, 'public.' + key :
+            if k in self :
+                return self [k]
+            if isinstance (k, ustr) :
+                k = k.encode ('utf-8')
+                if k in self :
+                    return self [k]
         raise AttributeError (key)
     # end def __getattr__
 
+    def __setattr__ (self, key, value) :
+        if isinstance (key, ustr) :
+            key = key.encode ('utf-8')
+        self [key] = value
+    # end def __setattr__
+
     def set_done (self, done = True) :
         self.done = done
     # end def set_done
@@ -486,18 +515,12 @@ class ACL (autosuper) :
         for (name, schema, frm) in self.revoke :
             r.append \
                 ( b'REVOKE %s ON SCHEMA %s FROM %s;'
-                % ( name.encode ('utf-8')
-                  , schema.encode ('utf-8')
-                  , frm.encode ('utf-8')
-                  )
+                % (name, schema, frm)
                 )
         for (name, schema, to) in self.grant :
             r.append \
                 ( b'GRANT %s ON SCHEMA %s TO %s;'
-                % ( name.encode ('utf-8')
-                  , schema.encode ('utf-8')
-                  , to.encode ('utf-8')
-                  )
+                % (name, schema, to)
                 )
         if self.grant or self.revoke :
             r.append (b'')
@@ -508,29 +531,48 @@ class ACL (autosuper) :
 # end class ACL
 
 class Name_Item (autosuper) :
-    name_re = re.compile ('[^a-zA-Z0-9_]')
+    name_re = re.compile (b'[^a-zA-Z0-9_]')
 
     # Probably lots of keywords missing
     keywords = dict.fromkeys \
-        (('time', 'order', 'position', 'interval', 'null', 'update'))
+        ((b'time', b'order', b'position', b'interval', b'null', b'update'))
 
     def __init__ (self, name) :
+        self.with_schema = False
         self.name = name
         self.__super.__init__ ()
     # end def __init__
 
     @property
     def quoted_name (self) :
-        return b"'%s'" % self.name.encode ('utf-8')
+        return b"'%s'" % self.name
     # end def quoted_name
 
+    @property
+    def quoted_fullname (self) :
+        n = self.name
+        if self.with_schema :
+            n = b'.'.join ((self.schema, self.name))
+        return b"'%s'" % n
+    # end def quoted_fullname
+
     @property
     def formatted_name (self) :
+        n = self.name
+        if self.with_schema :
+            n = b'.'.join ((self.schema, self.name))
         if self.name in self.keywords or self.name_re.search (self.name) :
-            return b'"%s"' % self.name.encode ('utf-8')
-        return self.name.encode ('utf-8')
+            return b'"%s"' % n
+        return n
     # end def formatted_name
 
+    @property
+    def short_name (self) :
+        if self.name in self.keywords or self.name_re.search (self.name) :
+            return b'"%s"' % self.name
+        return self.name
+    # end def short_name
+
 # end class Name_Item
 
 class Column (Name_Item) :
@@ -540,21 +582,25 @@ class Column (Name_Item) :
         self.typecl    = typecl
         self.nullable  = nullable
         self.sequences = []
+        self.seq_names = {}
         self.default   = None
         self.__super.__init__ (name)
     # end def __init__
 
     def append_sequence (self, seq) :
+        if seq.name in self.seq_names :
+            return
         self.sequences.append (seq)
+        self.seq_names [seq.name] = seq
         seq.column = self
     # end def append_sequence
 
     def as_pgsql (self) :
         r = [self.formatted_name]
-        r.append (self.typecl.typ (self.typ).encode ('utf-8'))
+        r.append (self.typecl.typ (self.typ))
         if self.default :
             r.append (b'DEFAULT')
-            r.append (self.default.encode ('utf-8'))
+            r.append (self.default)
         if not self.nullable :
             r.append (b'NOT NULL')
         return b' '.join (r)
@@ -575,7 +621,7 @@ class Extension (Name_Item) :
         r.append (b'')
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s' % self.short_name)
         x.append (b'Type: EXTENSION')
         x.append (b'Schema: -')
         x.append (b'Owner: ')
@@ -584,14 +630,14 @@ class Extension (Name_Item) :
         r.append (b'')
         r.append \
             ( b'CREATE EXTENSION IF NOT EXISTS %s WITH SCHEMA %s;'
-            % (self.formatted_name, self.schema.encode ('utf-8'))
+            % (self.formatted_name, self.schema)
             )
         r.append (b'')
         r.append (b'')
         if self.type :
             r.append (b'--')
             x = []
-            x.append (b'-- Name: EXTENSION %s' % self.formatted_name)
+            x.append (b'-- Name: EXTENSION %s' % self.short_name)
             x.append (b'Type: COMMENT')
             x.append (b'Schema: -')
             x.append (b'Owner: ')
@@ -600,7 +646,7 @@ class Extension (Name_Item) :
             r.append (b'')
             r.append \
                 ( b"COMMENT ON EXTENSION %s IS '%s';"
-                % (self.formatted_name, self.type.encode ('utf-8'))
+                % (self.formatted_name, self.type)
                 )
             r.append (b'')
             r.append (b'')
@@ -621,7 +667,7 @@ class Foreign_Key (Name_Item) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s' % self.short_name)
         x.append (b'Type: FK CONSTRAINT')
         x.append (b'Schema: %s' % self.table.schema.encode ('utf-8'))
         x.append (b'Owner: %s' % self.table.owner.encode ('utf-8'))
@@ -648,9 +694,10 @@ class Foreign_Key (Name_Item) :
 class Index (Name_Item) :
 
     def __init__ (self, name, typ, columns, ops) :
-        self.typ     = typ
-        self.columns = columns
-        self.ops     = ops
+        self.typ       = typ
+        self.columns   = columns
+        self.ops       = ops
+        self.is_unique = False
         self.__super.__init__ (name)
     # end def __init__
 
@@ -658,19 +705,24 @@ class Index (Name_Item) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s' % self.short_name)
         x.append (b'Type: INDEX')
-        x.append (b'Schema: %s' % self.table.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.table.owner.encode ('utf-8'))
-        x.append (b'Tablespace: %s' % self.table.tspace.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.table.schema)
+        x.append (b'Owner: %s' % self.table.owner)
+        if self.table.tspace :
+            x.append (b'Tablespace: %s' % self.table.tspace)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
+        u = b''
+        if self.is_unique :
+            u = b'UNIQUE '
         r.append \
-            ( b'CREATE INDEX %s ON %s USING %s (%s);'
-            % ( self.formatted_name
+            ( b'CREATE %sINDEX %s ON %s USING %s (%s);'
+            % ( u
+              , self.formatted_name
               , self.table.formatted_name
-              , self.typ.encode ('utf-8')
+              , self.typ
               , b', '.join (self.col_as_pgsql (c) for c in self.columns)
               )
             )
@@ -681,7 +733,7 @@ class Index (Name_Item) :
 
     def col_as_pgsql (self, column) :
         name = column.name
-        op   = self.ops.get (name, b'').encode ('utf-8')
+        op   = self.ops.get (name, b'')
         if op :
             return b' '.join ((column.formatted_name, op))
         return column.formatted_name
@@ -703,11 +755,12 @@ class Key (Name_Item) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s %s' % (self.table.short_name, self.short_name))
         x.append (b'Type: CONSTRAINT')
-        x.append (b'Schema: %s' % self.table.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.table.owner.encode ('utf-8'))
-        x.append (b'Tablespace: %s' % self.table.tspace.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.table.schema)
+        x.append (b'Owner: %s' % self.table.owner)
+        if self.table.tspace :
+            x.append (b'Tablespace: %s' % self.table.tspace)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
@@ -715,7 +768,7 @@ class Key (Name_Item) :
         r.append \
             ( b'    ADD CONSTRAINT %s %s (%s);'
             % ( self.formatted_name
-              , self.typ.encode ('utf-8')
+              , self.typ
               , b', '.join (k.formatted_name for k in self.columns)
               )
             )
@@ -735,10 +788,7 @@ class Set_Statement (Name_Item) :
 
     def as_pgsql (self) :
         r = []
-        r.append \
-            ( b'SET %s = %s;'
-            % (self.formatted_name, self.value.encode ('utf-8'))
-            )
+        r.append (b'SET %s = %s;' % (self.formatted_name, self.value))
         return b'\n'.join (r)
     # end def as_pgsql
 
@@ -750,6 +800,10 @@ class DB_Object (Name_Item) :
         self.owner  = ''
         self.schema = ''
         self.__super.__init__ (name)
+        if b'.' in self.name :
+            # Should only contain a single dot
+            self.schema, self.name = self.name.split (b'.')
+            self.with_schema = True
     # end def __init__
 
     def set_schema (self, schema) :
@@ -771,10 +825,7 @@ class DB_Object (Name_Item) :
         if self.owner :
             r.append \
                 ( b'ALTER TABLE %s.%s OWNER TO %s;'
-                % ( self.schema.encode ('utf-8')
-                  , self.formatted_name
-                  , self.owner.encode ('utf-8')
-                  )
+                % (self.schema, self.name, self.owner)
                 )
             r.append (b'')
         return b'\n'.join (r)
@@ -799,40 +850,41 @@ class Sequence (DB_Object) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s' % self.short_name)
         x.append (b'Type: SEQUENCE')
-        x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.schema)
+        x.append (b'Owner: %s' % self.owner)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
         r.append (b'CREATE SEQUENCE %s' % self.formatted_name)
-        r.append (b'    START WITH %s' % self.start)
-        r.append (b'    INCREMENT BY %s' % self.inc)
+        r.append (b'    START WITH %d' % self.start)
+        r.append (b'    INCREMENT BY %d' % self.inc)
         if self.min is None :
             r.append (b'    NO MINVALUE')
         if self.max is None :
             r.append (b'    NO MAXVALUE')
-        r.append (b'    CACHE %s;' % self.cache)
-        r.append (b'')
+        r.append (b'    CACHE %d;' % self.cache)
         r.append (b'')
         r.append (self.owner_as_pgsql ())
-        r.append (b'--')
-        x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
-        x.append (b'Type: SEQUENCE OWNED BY')
-        x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
-        r.append (b'; '.join (x))
-        r.append (b'--')
-        r.append (b'')
-        col = self.column.name.encode ('utf-8')
-        tbl = self.column.table.name.encode ('utf-8')
-        r.append \
-            ( b'ALTER SEQUENCE %s OWNED BY %s.%s;'
-            % (self.formatted_name, tbl, col)
-            )
-        r.append (b'')
+        if getattr (self, 'column', None) :
+            r.append (b'')
+            r.append (b'--')
+            x = []
+            x.append (b'-- Name: %s' % self.short_name)
+            x.append (b'Type: SEQUENCE OWNED BY')
+            x.append (b'Schema: %s' % self.schema)
+            x.append (b'Owner: %s' % self.owner)
+            r.append (b'; '.join (x))
+            r.append (b'--')
+            r.append (b'')
+            col = self.column.name
+            tbl = self.column.table
+            r.append \
+                ( b'ALTER SEQUENCE %s OWNED BY %s.%s;'
+                % (self.formatted_name, tbl, col)
+                )
+            r.append (b'')
         r.append (b'')
         return b'\n'.join (r)
     # end def as_pgsql
@@ -841,10 +893,10 @@ class Sequence (DB_Object) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.column.formatted_name)
+        x.append (b'-- Name: %s' % self.column.short_name)
         x.append (b'Type: DEFAULT')
-        x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.schema)
+        x.append (b'Owner: %s' % self.owner)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
@@ -860,14 +912,36 @@ class Sequence (DB_Object) :
         r.append (b'')
         return b'\n'.join (r)
     # end def default_as_pgsql
+
+    def init_as_pgsql (self) :
+        r = []
+        r.append (b'--')
+        x = []
+        x.append (b'-- Name: %s' % self.short_name)
+        x.append (b'Type: SEQUENCE SET')
+        x.append (b'Schema: %s' % self.schema)
+        x.append (b'Owner: %s' % self.owner)
+        r.append (b'; '.join (x))
+        r.append (b'--')
+        r.append (b'')
+        tf = [b'false', b'true'][self.booly]
+        r.append \
+            ( b'SELECT pg_catalog.setval(%s, %d, %s);'
+            % (self.quoted_fullname, self.value, tf)
+            )
+        r.append (b'')
+        r.append (b'')
+        return b'\n'.join (r)
+    # end def init_as_pgsql
+
 # end class Sequence
 
 class Table (DB_Object) :
 
     def __init__ (self, name) :
-        self.owner        = ''
-        self.schema       = ''
-        self.tspace       = ''
+        self.owner        = b''
+        self.schema       = b''
+        self.tspace       = b''
         self.columns      = []
         self.by_col       = {}
         self.contents     = []
@@ -893,7 +967,7 @@ class Table (DB_Object) :
     def append_index (self, idx) :
         self.indeces.append (idx)
         idx.table = self
-    # end def append_key
+    # end def append_index
 
     def append_key (self, key) :
         self.keys.append (key)
@@ -911,11 +985,12 @@ class Table (DB_Object) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Name: %s' % self.formatted_name)
+        x.append (b'-- Name: %s' % self.short_name)
         x.append (b'Type: TABLE')
-        x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
-        x.append (b'Tablespace: %s' % self.tspace.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.schema)
+        x.append (b'Owner: %s' % self.owner)
+        if self.tspace :
+            x.append (b'Tablespace: %s' % self.tspace)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
@@ -947,10 +1022,10 @@ class Table (DB_Object) :
         r = []
         r.append (b'--')
         x = []
-        x.append (b'-- Data for Name: %s' % self.formatted_name)
+        x.append (b'-- Data for Name: %s' % self.short_name)
         x.append (b'Type: TABLE DATA')
-        x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-        x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
+        x.append (b'Schema: %s' % self.schema)
+        x.append (b'Owner: %s' % self.owner)
         r.append (b'; '.join (x))
         r.append (b'--')
         r.append (b'')
@@ -964,6 +1039,8 @@ class Table (DB_Object) :
             for col in self.columns :
                 cn = col.name
                 v = line [cn]
+                if isinstance (v, string_types) :
+                    v = v.encode ('utf-8')
                 c.append (col.typecl.format (dialect_pg, col.typ, v))
             r.append (b'\t'.join (c))
         r.append (b'\\.')
@@ -971,22 +1048,7 @@ class Table (DB_Object) :
         r.append (b'')
         for col in self.columns :
             for seq in col.sequences :
-                r.append (b'--')
-                x = []
-                x.append (b'-- Name: %s' % seq.formatted_name)
-                x.append (b'Type: SEQUENCE SET')
-                x.append (b'Schema: %s' % self.schema.encode ('utf-8'))
-                x.append (b'Owner: %s' % self.owner.encode ('utf-8'))
-                r.append (b'; '.join (x))
-                r.append (b'--')
-                r.append (b'')
-                tf = [b'false', b'true'][seq.booly]
-                r.append \
-                    ( b'SELECT pg_catalog.setval(%s, %s, %s);'
-                    % (seq.quoted_name, seq.value, tf)
-                    )
-                r.append (b'')
-                r.append (b'')
+                r.append (seq.init_as_pgsql ())
         return b'\n'.join (r)
     # end def content_as_pgsql
 
@@ -1040,6 +1102,9 @@ class SQL_Parser (Parser) :
     re_index   = re.compile (br'CREATE INDEX (\S+) ON (\S+) USING (\S+) '
                              br'\(([^)]+)\);'
                             )
+    re_uindex  = re.compile (br'CREATE UNIQUE INDEX (\S+) ON (\S+) USING (\S+) '
+                             br'\(([^)]+)\);'
+                            )
     re_set     = re.compile (br'SET (\S+) = ([^;]+);')
     re_ext     = re.compile (br'CREATE EXTENSION IF NOT EXISTS (\S+) '
                              br'WITH SCHEMA (\S+);'
@@ -1061,6 +1126,7 @@ class SQL_Parser (Parser) :
         , ["init",  re_sq_set,  "init",  "seq_setval"]
         , ["init",  re_ato,     "cons",  "cons_start"]
         , ["init",  re_index,   "init",  "create_index"]
+        , ["init",  re_uindex,  "init",  "create_uindex"]
         , ["init",  re_set,     "init",  "set_stmt"]
         , ["init",  re_ext,     "init",  "ext_stmt"]
         , ["init",  re_comment, "init",  "ext_comment"]
@@ -1084,19 +1150,35 @@ class SQL_Parser (Parser) :
         ]
 
     def __init__ (self, fix_double_encode = False, *args, **kw) :
-        self.tables            = {}
+        self.tables            = adict ()
         self.tablenames        = []
-        self.sequences         = {}
+        self.sequences         = adict ()
+        self.free_seq          = {}
         self.fix_double_encode = fix_double_encode
         self.objects           = []
         self.extensions        = {}
         self.acl               = ACL ()
+        self.droptable         = {}
+        self.emptytable        = {}
+        self.cb                = {}
         # remember column names in creation order, should really use an
         # OrderedDict in self.tables but this means at least python2.7
         self.columns  = {}
         self.__super.__init__ (*args, **kw)
     # end def __init__
 
+    def register_empty_table (self, tablename) :
+        self.emptytable [tablename.encode ('utf-8')] = True
+    # end def register_empty_table
+
+    def register_drop_table (self, tablename) :
+        self.droptable [tablename.encode ('utf-8')] = True
+    # end def register_drop_table
+
+    def register_table_callback (self, tablename, method) :
+        self.cb [tablename.encode ('utf-8')] = method
+    # end def register_table_callback
+
     def as_pgsql (self) :
         r = []
         r.append (b'--')
@@ -1113,9 +1195,15 @@ class SQL_Parser (Parser) :
             df = tbl.seq_defaults ()
             if df :
                 r.append (df)
+        for sn in sorted (self.free_seq) :
+            seq = self.free_seq [sn]
+            r.append (seq.as_pgsql ())
         for tn in self.tablenames :
             tbl = self.tables [tn]
             r.append (tbl.content_as_pgsql ())
+        for sn in sorted (self.free_seq) :
+            seq = self.free_seq [sn]
+            r.append (seq.init_as_pgsql ())
         keys = []
         for tn in self.tablenames :
             tbl  = self.tables [tn]
@@ -1150,7 +1238,7 @@ class SQL_Parser (Parser) :
     def cons_end (self, state, new_state, match) :
         name = match.group (1)
         typ  = match.group (2)
-        cns  = match.group (3).split (',')
+        cns  = match.group (3).split (b',')
         cols = []
         for c in cns :
             c = c.strip ()
@@ -1163,41 +1251,64 @@ class SQL_Parser (Parser) :
         line    = self.line.rstrip (b'\n')
         tbl     = self.table
         fields  = self.fields
+        if tbl is None or fields is None :
+            return
         dfields = line.split (b'\t')
         # compensate for rstrip
-        for x in xrange (len (fields) - len (dfields)) :
+        for x in range (len (fields) - len (dfields)) :
             dfields.append (b'')
-        tbl.contents.append \
-            (adict ((a, tbl [a].typecl (b)) for a, b in zip (fields, dfields)))
+        contents = adict \
+            ((a, tbl [a].typecl (b)) for a, b in zip (fields, dfields))
+        doit = True
+        if tbl.name in self.cb :
+            doit = self.cb [tbl.name](contents)
+        if doit :
+            tbl.contents.append (contents)
     # end def copy_entry
 
     def copy_start (self, state, new_state, match) :
-        self.tablename = match.group (1)
-        self.table     = self.tables [self.tablename]
-        self.fields    = [x.strip ('"') for x in match.group (2).split (', ')]
+        name  = match.group (1)
+        table = self.tables.get (name)
+        if not table :
+            self.table = self.tablename = self.fields = None
+            return
+        if table.name in self.droptable or table.name in self.emptytable :
+            self.table = self.tablename = self.fields = None
+            return
+        self.tablename = name
+        self.table     = table
+        self.fields    = [x.strip (b'"') for x in match.group (2).split (b', ')]
     # end def copy_start
 
     def create_index (self, state, new_state, match) :
         name = match.group (1)
-        tbl  = self.tables [match.group (2)]
+        tbl  = self.tables.get (match.group (2))
+        if not tbl :
+            return
         typ  = match.group (3)
-        cns  = match.group (4).split (',')
+        cns  = match.group (4).split (b',')
         cols = []
         ops  = {}
         for c in cns :
             c = c.strip ()
-            cops = c.split (' ', 1)
-            op  = ''
+            cops = c.split (b' ', 1)
+            op  = b''
             if len (cops) > 1 :
                 c = cops [0]
                 op = cops [1]
                 ops [c] = op
-            c = c.strip ('"')
+            c = c.strip (b'"')
             col = tbl [c]
             cols.append (col)
         tbl.append_index (Index (name, typ, cols, ops))
     # end def create_index
 
+    def create_uindex (self, state, new_state, match) :
+        r = self.create_index (state, new_state, match)
+        tbl = self.tables [match.group (2)]
+        tbl.indeces [-1].is_unique = True
+    # end def create_uindex
+
     def dump (self) :
         for tn in self.tablenames :
             tbl = self.tables [tn]
@@ -1241,10 +1352,10 @@ class SQL_Parser (Parser) :
     def insert (self, state, new_state, match) :
         """ This asumes the whole insert statement is one line. """
         name   = match.group (1)
-        tpl    = match.group (2).split ('),(')
+        tpl    = match.group (2).split (b'),(')
         tuples = []
         for t in tpl :
-            tuples.append (t.replace ('\\n', '\n').replace ('\\r', '\r'))
+            tuples.append (t.replace (b'\\n', b'\n').replace (b'\\r', b'\r'))
         tbl    = self.tables [name]
         fields = tbl.columns
         reader = csv.reader \
@@ -1261,14 +1372,16 @@ class SQL_Parser (Parser) :
         schema = ''
         name   = match.group (1)
         owner  = match.group (2)
-        if '.' in name :
-            schema, name = name.rsplit ('.', 1)
-        if name in self.tables :
-            obj = self.tables [name]
-        else :
-            obj = self.sequences [name]
-        obj.set_owner (owner)
-        obj.set_schema (schema)
+        ename  = name
+        if b'.' in name :
+            schema, ename = name.rsplit (b'.', 1)
+        for n in ename, name :
+            for d in self.tables, self.sequences :
+                obj = d.get (n, None)
+                if obj is not None :
+                    obj.set_owner (owner)
+                    obj.set_schema (schema)
+                    return
     # end def owner
 
     def revoke_stmt (self, state, new_state, match) :
@@ -1282,6 +1395,7 @@ class SQL_Parser (Parser) :
         tbl = self.tables [tbln]
         col = tbl [coln]
         col.append_sequence (seq)
+        del self.free_seq [name]
     # end def seq_alter
 
     def seq_cache (self, state, new_state, match) :
@@ -1304,13 +1418,14 @@ class SQL_Parser (Parser) :
         name = match.group (1)
         seq  = self.sequences [name]
         seq.value = int (match.group (2))
-        seq.booly = (match.group (3) == 'true')
+        seq.booly = (match.group (3) == b'true')
     # end def seq_inc
 
     def seq_start (self, state, new_state, match) :
         name = match.group (1)
         self.seq = Sequence (name)
         self.sequences [name] = self.seq
+        self.free_seq  [name] = self.seq
     # end def seq_start
 
     def seq_startwith (self, state, new_state, match) :
@@ -1329,63 +1444,68 @@ class SQL_Parser (Parser) :
             But the dump is in utf-8 anyway (!)
         """
         m = self.re_charset.search (self.line)
-        for c in self.table.columns :
-            if 0 and m :
-                c.charset = m.group (1)
-            if self.fix_double_encode :
-                c.typecl.fix_double_encode = True
+        if self.table is not None :
+            for c in self.table.columns :
+                if 0 and m :
+                    c.charset = m.group (1)
+                if self.fix_double_encode :
+                    c.typecl.fix_double_encode = True
         self.table = None
     # end def table_end
 
     def table_entry (self, state, new_state, match) :
+        if self.table is None :
+            return
         line = self.line.strip ()
         pars = ()
         try :
             name, typ, rest = line.split (None, 2)
         except ValueError :
             name, typ = line.split (None, 1)
-            if typ.endswith (',') :
+            if typ.endswith (b',') :
                 typ = typ [:-1]
-            rest = ''
-        typ  = typ.rstrip (',')
-        rest = rest.rstrip (',')
-        if '(' in typ :
-            pars = typ [typ.index ('(') + 1:typ.index (')')].split (',')
-        elif '(' in rest :
-            pars = rest [rest.index ('(') + 1:rest.index (')')].split (',')
-        if name.startswith ('"') or name.startswith ('`') :
+            rest = b''
+        typ  = typ.rstrip (b',')
+        rest = rest.rstrip (b',')
+        if b'(' in typ :
+            pars = typ [typ.index (b'(') + 1:typ.index (b')')].split (b',')
+        elif b'(' in rest :
+            pars = rest [rest.index (b'(') + 1:rest.index (b')')].split (b',')
+        if name.startswith (b'"') or name.startswith (b'`') :
             name = name [1:-1]
-        tn = typ.split ('(', 1) [0]
-        if tn == 'character' and rest.startswith ('varying') :
-            tn = 'character varying'
-        if rest.startswith ('with time zone') :
-            tn = tn + ' with time zone'
-        if rest.startswith ('without time zone') :
-            tn = tn + ' without time zone'
-        if typ.startswith ('int(') or typ.startswith ('tinyint(') :
-            typ = 'integer'
-        if typ.startswith ('varchar') :
-            typ = 'varchar'
-        if typ.startswith ('enum') :
-            typ = 'enum'
-        if typ.startswith ('numeric') :
-            typ = 'numeric'
-        if name in ('PRIMARY', 'UNIQUE') and typ == 'KEY' :
+        tn = typ.split (b'(', 1) [0]
+        if tn == b'character' and rest.startswith (b'varying') :
+            tn = b'character varying'
+        if rest.startswith (b'with time zone') :
+            tn = tn + b' with time zone'
+        if rest.startswith (b'without time zone') :
+            tn = tn + b' without time zone'
+        if typ.startswith (b'int(') or typ.startswith (b'tinyint(') :
+            typ = b'integer'
+        if typ.startswith (b'varchar') :
+            typ = b'varchar'
+        if typ.startswith (b'enum') :
+            typ = b'enum'
+        if typ.startswith (b'numeric') :
+            typ = b'numeric'
+        if name in (b'PRIMARY', b'UNIQUE') and typ == b'KEY' :
             # FIXME: Should produce a key object
             self.table.append_key (line)
             return
-        if name == 'KEY' :
+        if name == b'KEY' :
             # FIXME: Should produce a key object
             self.table.append_key (line)
             return
-        if name == 'CONSTRAINT' :
+        if name == b'CONSTRAINT' :
             self.table.append_constraint (line)
             return
-        method = getattr (self, 'type_' + typ, self.type_default)
+        t = typ.decode ('ascii')
+        r = rest.decode ('ascii')
+        method = getattr (self, 'type_' + t, self.type_default)
         nullable = True
-        if rest.endswith ('NOT NULL') :
+        if rest.endswith (b'NOT NULL') :
             nullable = False
-        col = Column (name, tn, method (typ, rest, pars), nullable)
+        col = Column (name, tn, method (t, r, pars), nullable)
         self.table.append_column (col)
         m = self.re_default.search (rest)
         if m :
@@ -1393,9 +1513,13 @@ class SQL_Parser (Parser) :
     # end def table_entry
 
     def table_start (self, state, new_state, match) :
-        name = match.group (1).strip ('`')
-        self.table = self.tables  [name] = Table (name)
-        self.tablenames.append (name)
+        name = match.group (1).strip (b'`')
+        tbl  = Table (name)
+        if tbl.name in self.droptable :
+            self.table = None
+        else :
+            self.table = self.tables [name] = tbl
+            self.tablenames.append (name)
     # end def table_start
 
     # Magic type methods for SQL types:
@@ -1453,5 +1577,5 @@ if __name__ == "__main__" :
         f = sys.stdin
     c = SQL_Parser ()
     c.parse    (f)
-    print (c.as_pgsql ())
+    print (c.as_pgsql ().decode ('utf-8'))
 ### __END__ sqlparser
