@@ -1,6 +1,27 @@
-#!/usr/bin/python
+#!/usr/bin/python3
+# Copyright (C) 2010-21 Dr. Ralf Schlatterbeck Open Source Consulting.
+# Reichergasse 131, A-3411 Weidling.
+# Web: http://www.runtux.com Email: office@runtux.com
+# All rights reserved
+# ****************************************************************************
+# This library is free software; you can redistribute it and/or modify
+# it under the terms of the GNU Library General Public License as
+# published by the Free Software Foundation; either version 2 of the
+# License, or (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Library General Public License for more details.
+#
+# You should have received a copy of the GNU Library General Public
+# License along with this program; if not, write to the Free Software
+# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+# ****************************************************************************
 
+from __future__ import print_function
 from operator          import or_
+from functools         import reduce
 from rsclib.autosuper  import autosuper
 from rsclib.execute    import Exec
 from rsclib.IP_Address import IP4_Address
@@ -829,6 +850,7 @@ if __name__ == '__main__' :
     shaper   = Shaper ('/bin/tc', root)
 
     for bw, dev in (2000, 'eth0'), (1000, 'ifb0=eth0') :
-        print >> sys.stdout, shaper.generate \
-            (bw, dev, rulefile = open ('prerouting.out'))
-
+        print \
+            ( shaper.generate (bw, dev, rulefile = open ('prerouting.out'))
+            , file = sys.stdout
+            )
